import numpy as np
import matplotlib.pyplot as plt

theta = 0.25 * np.pi
"""
ndata = 50
sqrtV = 0.4
x0 = np.sort(10*np.random.rand(ndata))
y0 = sqrtV * np.random.randn(ndata)
vecs = np.array([x0, y0])
rotMatrix = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
vecsR = np.dot(rotMatrix, vecs)
x1 = vecsR[0, :]
y1 = vecsR[1, :]
plt.plot(x0, y0, linestyle="none", marker=".")
plt.plot(x1, y1, linestyle="none", marker=".")
plt.show()
"""

print np.log(np.e)
print np.cos(np.pi/3)
